from datetime import datetime
import fnmatch
import itertools
import os
from pathlib import Path
import random
import re
import shutil
import string
import urllib.parse
import zipfile
import boto3
import yaml

success_response = {'statusCode': 200}


class WildcardDict(dict):

    def get_matching(self, file_name, default=None):
        return next((self[key] for key in self if fnmatch.fnmatch(file_name, key)), default)


def lambda_handler(event, context):

    try:
        s3_record = event['Records'][0]['s3']
        bucket_name = s3_record['bucket']['name']
        object_key = s3_record['object']['key']
        file_name = urllib.parse.unquote_plus(object_key, encoding='utf-8')
    except Exception as e:
        print(f'Error parsing the S3 event: {repr(e)}')
        raise e

    # We are looking for the YAML configuration file as synchronization trigger,
    # so we ignore all other files
    if not file_name.lower().endswith(('.yml', '.yaml')):
        return success_response

    print(f'Processing YAML configuration file "{file_name}"...')

    try:
        s3_client = boto3.client('s3')
        yaml_object = s3_client.get_object(Bucket=bucket_name, Key=file_name)
    except Exception as e:
        print(f'Error getting object "{file_name}" from bucket "{bucket_name}": {repr(e)}')
        raise e

    try:
        yaml_config = yaml.safe_load(yaml_object['Body'])
        repo_config = yaml_config['repository-config']
        repo_name = repo_config['target-repository']
        branch_name = repo_config.get('target-branch', 'main')
        commit_message = repo_config.get(
            'commit-message',
            f'Synchronize sources at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}')

        files_lookup = WildcardDict({'*.yaml': '/', '*.yml': '/'})

        _walk_folder_structure(files_lookup, repo_config['directory-structure'], '/')

    except Exception as e:
        print(f'Error parsing the config YAML: {repr(e)}')
        raise e

    code_commit_client = boto3.client('codecommit')

    try:
        branch_info = code_commit_client.get_branch(
            repositoryName=repo_name, branchName=branch_name)

        last_commit = branch_info['branch']['commitId']
    except code_commit_client.exceptions.BranchDoesNotExistException:
        print('Branch does not exist')
        last_commit = None
    except code_commit_client.exceptions.RepositoryDoesNotExistException:
        print('Repository does not exist, creating one')
        last_commit = None
        pipeline_config = yaml_config.get('pipeline-config', None)
        _create_repo_and_pipeline(code_commit_client, repo_name, branch_name, pipeline_config)

    s3 = boto3.resource('s3')
    source_bucket = s3.Bucket(bucket_name)

    print(f'Pushing files to repo "{repo_name}" and branch "{branch_name}"')

    try:
        all_objects = source_bucket.objects.all()
        items_to_commit = _get_files_to_commit(all_objects, files_lookup, s3_client, bucket_name)

        commit_args = {
            'repositoryName': repo_name,
            'branchName': branch_name,
            'authorName': 'Automation',
            'commitMessage': commit_message
        }

        _chunk_and_push_items(items_to_commit, code_commit_client, commit_args, last_commit)
        shutil.rmtree('/tmp/zips', ignore_errors=True)

    except Exception as e:
        print(f'Error committing files: {repr(e)}')
        raise

    try:
        source_bucket.objects.all().delete()
    except Exception as e:
        print(f'Error deleting the bucket contents: {repr(e)}')
        raise e

    return success_response


def _get_files_to_commit(s3_objects, files_lookup, s3_client, bucket_name):
    for s3_object in s3_objects:
        object_key = s3_object.key
        if object_key.lower().endswith('.zip'):
            print(f'Processing ZIP package "{object_key}"')
            files_from_zip = _get_files_from_zip(files_lookup, s3_client, bucket_name, object_key)
            for source_file_path, target_file_path in files_from_zip:
                content = _get_file_content(source_file_path)
                yield _get_commit_item(target_file_path, content)
        else:
            target_file_path = _get_target_file_path(files_lookup, object_key)
            if target_file_path is None:
                continue

            content = _get_bucket_object_content(s3_client, bucket_name, object_key)
            yield _get_commit_item(target_file_path, content)


def _get_commit_item(file_path, content):
    return {'filePath': file_path, 'fileContent': content}


def _get_files_from_zip(files_lookup, s3_client, bucket_name, zip_package_key):
    zip_content = _get_bucket_object_content(s3_client, bucket_name, zip_package_key)
    zip_file_name = os.path.join('/tmp/zips', zip_package_key)
    os.makedirs('/tmp/zips', exist_ok=True)
    with open(zip_file_name, 'wb') as f:
        f.write(zip_content)

    try:
        zip_target_dir = zip_file_name + '_'
        with zipfile.ZipFile(zip_file_name, 'r') as zip_package:
            zip_package.extractall(zip_target_dir)

        for file_path in Path(zip_target_dir).rglob('*'):
            target_file_path = _get_target_file_path(files_lookup, file_path)
            if target_file_path is not None:
                yield (file_path, target_file_path)
    finally:
        os.remove(zip_file_name)
        shutil.rmtree(zip_target_dir, ignore_errors=True)


def _get_target_file_path(files_lookup, source_file_path):
    target_file_name = os.path.basename(source_file_path)
    matching_file_name = target_file_name.lower()

    target_path = files_lookup.get_matching(matching_file_name)

    if target_path is not None:
        target_path = os.path.join(target_path, target_file_name)
        print(f'File "{source_file_path}" will be stored in "{target_path}"')

    return target_path


def _walk_folder_structure(mapping, directory_list, parent_directory):
    for directory_entry in directory_list:
        directory_name = next(iter(directory_entry.keys()))
        files_config = directory_entry[directory_name].get('files', None)
        subdirs = directory_entry[directory_name].get('subdirs', None)
        if directory_name != '/':
            directory_name = directory_name.lstrip('/')
            if re.search(r'[\\/*?<>|:]', directory_name):
                raise Exception(f'Invalid characters in directory name "{directory_name}"')
        full_directory_path = os.path.join(parent_directory, directory_name)
        if files_config is not None:
            for file_pattern in files_config:
                mapping[file_pattern.lower()] = full_directory_path
        if subdirs is not None:
            _walk_folder_structure(mapping, subdirs, full_directory_path)


def _chunk_items(iterable, size):
    it = iter(iterable)
    item = list(itertools.islice(it, size))
    while item:
        yield item
        item = list(itertools.islice(it, size))


def _get_bucket_object_content(s3_client, bucket_name, object_key):
    object = s3_client.get_object(Bucket=bucket_name, Key=object_key)
    return object['Body'].read()


def _get_file_content(file_path):
    with open(file_path, 'rb') as f:
        return f.read()


def _chunk_and_push_items(items_list, code_commit_client, commit_args, last_commit):
    try:
        commit_files = {}

        for item in items_list:
            # We need to use 100 files per chunk
            # because AWS CodeCommit API accepts at most 100 files per commit
            if (len(commit_files) == 100):
                last_commit = _push_items(
                    commit_files.values(), code_commit_client, commit_args, last_commit)
                commit_files.clear()

            # Using a dict to prevent multiple content items for the same file
            # Last content wins
            commit_files[item['filePath']] = item

        last_commit = _push_items(
            commit_files.values(), code_commit_client, commit_args, last_commit)

        return last_commit
    except Exception as e:
        print(f'Error pushing to repository: {repr(e)}')
        raise


def _push_items(items_list, code_commit_client, commit_args, last_commit):
    items_count = len(items_list)
    if items_count == 0:
        return last_commit

    print(f'Pushing {items_count} file(s)...')

    commit_args['putFiles'] = list(items_list)

    if last_commit:
        commit_args['parentCommitId'] = last_commit

    try:
        response = code_commit_client.create_commit(**commit_args)
        return response['commitId']
    except code_commit_client.exceptions.NoChangeException:
        # If there are no changes detected in the uploaded files,
        # we treat this as a valid use case
        return last_commit
    except Exception as e:
        print(f'Error pushing to repository: {repr(e)}')
        raise


def _create_repo_and_pipeline(code_commit_client, repo_name, branch_name, pipeline_config):
    try:
        code_commit_client.create_repository(repositoryName=repo_name)
        print(f'Creating CodeCommit repository: {repo_name}')
    except Exception as e:
        print(f'Error creating repository: {repr(e)}')
        raise

    pipeline_source_url = os.environ['PIPELINE_URL']
    ps_suffix = os.environ['PS_SUFFIX']
    role = os.environ['PIPELINE_DEPLOYMENT_ROLE_ARN']
    alert_notifications = [] if pipeline_config is None else pipeline_config.get('alert-notifications', [])
    approval_notifications = [] if pipeline_config is None else pipeline_config.get('approval-notifications', [])

    stack_parameters = [
        {'ParameterKey': 'RepositoryName', 'ParameterValue': repo_name},
        {'ParameterKey': 'RepositoryBranch', 'ParameterValue': branch_name},
        {'ParameterKey': 'ParameterStoreSuffix', 'ParameterValue': ps_suffix},
        {'ParameterKey': 'RetainEnvironment', 'ParameterValue': 'false'}
    ]

    if len(alert_notifications) > 0:
        stack_parameters.append(
            {'ParameterKey': 'AdminEmailAddress', 'ParameterValue': ','.join(alert_notifications)})
    if len(approval_notifications) > 0:
        stack_parameters.append(
            {'ParameterKey': 'ApprovalEmailAddress', 'ParameterValue': ','.join(approval_notifications)})

    stack_name = repo_name + '-' + _get_short_uuid()

    cloud_formation_client = boto3.client('cloudformation')
    try:
        print('Creating pipeline Cloudformation stack')
        cloud_formation_client.create_stack(
            StackName=stack_name,
            TemplateURL=pipeline_source_url,
            Parameters=stack_parameters,
            RoleARN=role,
            Capabilities=['CAPABILITY_NAMED_IAM'])
        waiter = cloud_formation_client.get_waiter('stack_create_complete')
        waiter.wait(StackName=stack_name)
        print(f'Pipeline stack ({stack_name}) creation complete')
    except Exception as e:
        print(f'Error deploying the pipeline: {repr(e)}')
        raise


def _get_short_uuid():
    alphabet = string.ascii_lowercase + string.digits
    return ''.join(random.choices(alphabet, k=8))
